/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_BANKING_RECONCILEDBANKACCOUNT_H_
#define DRN_BANKING_RECONCILEDBANKACCOUNT_H_

#include <pecunia/Money.h>

#include <QDate>

#include <iosfwd>
#include <optional>

#include <banking_export.h>
#include "BankAccount.h"


namespace pecunia::currency
{

enum class Currency : quint16;

}

namespace drn::accounting
{

class AccountNumber;

}
namespace drn::banking
{

class DRN_BANKING_EXPORT ReconciledBankAccount
{
	BankAccount bankAccount_;
	pecunia::currency::Money balance_;
	std::optional<::QDate> reconciledOn_;

public:
	ReconciledBankAccount() = default;
	ReconciledBankAccount(BankAccount bankAccount, const pecunia::currency::Currency& currency);
	ReconciledBankAccount(
		BankAccount bankAccount,
		pecunia::currency::Money balance,
		std::optional<::QDate> reconciledOn
	);
	ReconciledBankAccount(const ReconciledBankAccount& other) noexcept;
	ReconciledBankAccount(ReconciledBankAccount&& other) noexcept;
	~ReconciledBankAccount() noexcept;
	ReconciledBankAccount& operator=(const ReconciledBankAccount& other) noexcept;
	ReconciledBankAccount& operator=(ReconciledBankAccount&& other) noexcept;
	/**
	 * @brief An accessor to the bank account that is reconciled.
	 */
	[[nodiscard]]
	const BankAccount& bankAccount() const noexcept;
	/**
	 * @brief An accessor to the amount of money in the bank account on the date reconciled.
	 *
	 * @return When the bank account was reconciled, gives the reconciled balance, null when the
	 * bank account has yet to be reconciled.
	 */
	[[nodiscard]]
	const pecunia::currency::Money& balance() const noexcept;
	/**
	 * @brief An accessor to the date the bank account was reconciled on.
	 *
	 * @return When the bank account was reconciled, gives the date, null when the bank account has
	 * yet to be reconciled.
	 */
	[[nodiscard]]
	const std::optional<::QDate>& reconciledOn() const noexcept;
	/**
	 * @brief Updates the reconciled information.
	 *
	 * @exception BankError When the date for reconciliation is in the past of the current
	 * reconciliation.
	 *
	 * @param reconciledOn The date for the new balance after a successful reconciliation.
	 * @param balance The new balance after reconciliation.
	 */
	void reconcile(const ::QDate& reconciledOn, const pecunia::currency::Money& balance);
};

DRN_BANKING_EXPORT bool operator<(
	const ReconciledBankAccount& lhs,
	const ReconciledBankAccount& rhs
);
DRN_BANKING_EXPORT bool operator==(
	const ReconciledBankAccount& lhs,
	const ReconciledBankAccount& rhs
);
DRN_BANKING_EXPORT bool operator!=(
	const ReconciledBankAccount& lhs,
	const ReconciledBankAccount& rhs
);

DRN_BANKING_EXPORT std::ostream& operator<<(std::ostream& o, const ReconciledBankAccount& rba);

}

#endif
