/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_BUDGETING_ALGORITHM_H_
#define DRN_BUDGETING_ALGORITHM_H_

#include <QtGlobal>

#include <chrono>

#include "Constants.hpp"
#include <budgeting_export.h>


namespace pecunia::currency
{

class Money;

}

namespace drn::budgeting
{

enum class EventFrequency : quint8;

/**
 * @brief Accumulates a given amount of money per period over another period of time.
 *
 * @exception Error When either the supplied from or to values are Unknown.
 *
 * @param accumulatingAs The frequency period to accumulate an amount over that period of time.
 * @param amount The amount of money to be accumulated.
 * @param amountPeriod The period the amount of money is currently expressed in.
 * @param hrInWeek The number of hours considered in a full week.
 * @param dayInWeek The number of days considered in a full week.
 *
 * @return The accumulated amount as expressed in the new period.
 */
DRN_BUDGETING_EXPORT pecunia::currency::Money accumulateOver(
	const EventFrequency accumulatingAs,
	const pecunia::currency::Money& amount,
	const EventFrequency amountPeriod,
	const std::chrono::hours& hrInWeek = std::chrono::hours{maximumHoursInAWeek},
	const std::chrono::days& dayInWeek = std::chrono::days{maximumDaysInAWeek}
);
/**
 * @brief Adjusts a given amount of money per period to another period. When converting with
 * the "never" frequency, it is considered always to be a zero amount.
 *
 * @exception Error When either the supplied from or to values are Unknown.
 * @exception logic_error When either the from or to values are not known values.
 *
 * @param adjustingTo The frequency period to convert to.
 * @param amount The amount of money to convert.
 * @param amountPeriod The period the amount of money is currently expressed in.
 * @param hrInWeek The number of hours considered in a full week.
 * @param dayInWeek The number of days considered in a full week.
 *
 * @return The amount as expressed in the new period.
 */
DRN_BUDGETING_EXPORT pecunia::currency::Money adjustTo(
	const EventFrequency adjustingTo,
	const pecunia::currency::Money& amount,
	const EventFrequency amountPeriod,
	const std::chrono::hours& hrInWeek = std::chrono::hours{maximumHoursInAWeek},
	const std::chrono::days& dayInWeek = std::chrono::days{maximumDaysInAWeek}
);

}

#endif
