/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_FOUNDATION_OBSERVERPTR_HPP_
#define DRN_FOUNDATION_OBSERVERPTR_HPP_

#include <cassert>
#include <memory>
#include <ostream>


namespace drn::foundation
{

template <typename>
class UniqueQtPtr;

/**
 * @brief A smart pointer for observing another pointer. This means that it preforms no memory
 * management on the lifetime of the observed pointer. This functionally acts as a marker for other
 * developers that this pointer's does not need to be managed nor can it be accidentally deleted.
 *
 * @tparm ObservedType The type the smart pointer is observing.
 */
template<typename ObservedType>
class ObserverPtr
{
	ObservedType* observed_;

public:
	/// @brief The type the smart pointer is observing.
	using element_type = ObservedType;
	/// @brief The pointer type the smart pointer is observing.
	using pointer = ObservedType*;

	/**
	 * @brief Full initialisation constructor for observing a naked pointer.
	 *
	 * @param observe The pointer that is to be observed.
	 */
	ObserverPtr(ObservedType* const observe = nullptr);
	/**
	 * @brief Full initialisation constructor for observing a standard unique pointer.
	 *
	 * @param observe The pointer that is to be observed.
	 */
	ObserverPtr(const std::unique_ptr<ObservedType>& observe);
	/**
	 * @brief Full initialisation constructor for observing a Qt unique pointer.
	 *
	 * @param observe The pointer that is to be observed.
	 */
	ObserverPtr(const UniqueQtPtr<ObservedType>& observe);
	ObserverPtr(const ObserverPtr&) = default;
	ObserverPtr(ObserverPtr&&) = default;
	~ObserverPtr() = default;
	ObserverPtr& operator=(const ObserverPtr&) = default;
	ObserverPtr& operator=(ObserverPtr&&) = default;
	/**
	 * @brief An assignment operator for setting the object to observe another pointer.
	 *
	 * @param observe The pointer that is to be observed.
	 *
	 * @return The current object.
	 */
	ObserverPtr& operator=(ObservedType* const observe);
	/**
	 * @brief An assignment operator for setting the object to be null.
	 *
	 * @return The current object.
	 */
	ObserverPtr& operator=(const std::nullptr_t);
	/**
	 * @brief Equality operator for testing if this object points to null.
	 *
	 * @return True when the object it points to is null, false otherwise.
	 */
	bool operator==(const std::nullptr_t) const;
	/**
	 * @brief Equality operator for testing if this object points to another specified
	 * pointer.
	 *
	 * @param other The other observed pointer to compare against.
	 *
	 * @return True when both pointers point to the same address, false otherwise.
	 */
	bool operator==(const ObserverPtr& other) const;
	/**
	 * @brief Equality operator for testing if this object points to another specified
	 * pointer.
	 *
	 * @param other The other observed pointer to compare against.
	 *
	 * @return True when both pointers point to the same address, false otherwise.
	 */
	bool operator==(const ObservedType* observed) const;
	/**
	 * @brief Inequality operator for testing if this object points to null.
	 *
	 * @return True when the object it points to is not null, false otherwise.
	 */
	bool operator!=(const std::nullptr_t) const;
	/**
	 * @brief Inequality operator for testing if this object points to another specified
	 * pointer.
	 *
	 * @param other The other observed pointer to compare against.
	 *
	 * @return True when both pointers do not point to the same address, false otherwise.
	 */
	bool operator!=(const ObserverPtr& other) const;
	/**
	 * @brief Inequality operator for testing if this object points to another specified
	 * pointer.
	 *
	 * @param other The other observed pointer to compare against.
	 *
	 * @return True when both pointers do not point to the same address, false otherwise.
	 */
	bool operator!=(const ObservedType* observed) const;
	/**
	 * @brief Less-than operator for testing if the address observed by this object appears before
	 * another specified naked pointer's address.
	 *
	 * @param other The observed pointer to compare against.
	 *
	 * @return True when this object's address is less-than the address of the supplied pointer's
	 * address, false otherwise.
	 */
	bool operator<(const ObservedType* observed) const;
	/**
	 * @brief Less-than operator for testing if the address observed by this object appears before
	 * another specified pointer's address.
	 *
	 * @param other The observed pointer to compare against.
	 *
	 * @return True when this object's address is less-than the address of the supplied pointer's
	 * address, false otherwise.
	 */
	bool operator<(const ObserverPtr& other) const;
	/**
	 * @brief Dereference pointer to the observed pointer. The behaviour is undefined if the
	 * observed pointer is null.
	 *
	 * @return The pointer being observed.
	 */
	ObservedType& operator*();
	/**
	 * @brief Dereference pointer to the observed pointer. The behaviour is undefined if the
	 * observed pointer is null.
	 *
	 * @return The pointer being observed.
	 */
	ObservedType& operator*() const;
	/**
	 * @brief Dereference pointer to the observed pointer. The behaviour is undefined if the
	 * observed pointer is null.
	 *
	 * @return The pointer being observed.
	 */
	ObservedType* operator->();
	/**
	 * @brief Dereference pointer to the observed pointer. The behaviour is undefined if the
	 * observed pointer is null.
	 *
	 * @return The pointer being observed.
	 */
	ObservedType* operator->() const;
	/**
	 * @brief Determines if this object is observing a pointer.
	 *
	 * @return True when this object is not null, false otherwise.
	 */
	explicit operator bool() const noexcept;
	/**
	 * @brief Conversion operator to the underlying pointer being observed.
	 *
	 * @return The pointer being observed.
	 */
	explicit operator ObservedType*() const noexcept;
	void operator delete(void*) = delete;
	/**
	 * @brief An accessor to the underlying pointer being observed.
	 *
	 * @return The pointer being observed.
	 */
	ObservedType* get() const noexcept;
};

/**
 * @brief Equality operator test to determine a pointer that is being observed is null.
 *
 * @tparam ObservedType The type of the pointer being observed.
 *
 * @param rhs The object to test if it is null.
 *
 * @return True when the object is null, i.e. not observing anything, false otherwise.
 */
template<typename ObservedType>
bool operator==(std::nullptr_t, const ObserverPtr<ObservedType>& rhs);
/**
 * @brief Equality operator test to determine if a smart pointer is observing a specific naked
 * pointer.
 *
 * @tparam ObservedType The type of the naked pointer that is to be tested for being observed.
 * @tparam ObservedPtrType The type of the smart pointer that is to be tested for observing.
 *
 * @param lhs The naked pointer to test for being observed.
 * @param rhs The smart pointer to test for observing.
 *
 * @return True when the smart pointer points to the naked pointer, false otherwise.
 */
template<typename ObservedType, typename ObservedPtrType>
bool operator==(const ObservedType* lhs, const ObserverPtr<ObservedPtrType>& rhs);
/**
 * @brief Inequality operator test to determine a pointer that is being observed is null.
 *
 * @tparam ObservedType The type of the pointer being observed.
 *
 * @param rhs The object to test if it is null.
 *
 * @return True when the object is not null, i.e. observing something, false otherwise.
 */
template<typename ObservedType>
bool operator!=(std::nullptr_t, const ObserverPtr<ObservedType>& rhs);
/**
 * @brief Inequality operator test to determine if a smart pointer is observing a specific naked
 * pointer.
 *
 * @tparam ObservedType The type of the naked pointer that is to be tested for being observed.
 * @tparam ObservedPtrType The type of the smart pointer that is to be tested for observing.
 *
 * @param lhs The naked pointer to test for being observed.
 * @param rhs The smart pointer to test for observing.
 *
 * @return True when the smart pointer does not point to the naked pointer, false otherwise.
 */
template<typename ObservedType, typename ObservedPtrType>
bool operator!=(const ObservedType* lhs, const ObserverPtr<ObservedPtrType>& rhs);
/**
 * @brief Stream insertion operator for outputting the contests of the smart pointer. When the smart
 * pointer is observing another pointer, the object pointed to will have its stream insertion
 * operator used. When the smart pointer is null, the stream will be populated with "<nullptr>".
 *
 * @tparam ObservedType The type the smart pointer is observing.
 *
 * @param o The stream to place the contents into.
 * @param observer The smart pointer to output.
 *
 * @return The stream populated with the contents of the smart pointer.
 */
template<typename ObservedType>
std::ostream& operator<<(std::ostream& o, const ObserverPtr<ObservedType>& observer);

}

template<typename ObservedType>
drn::foundation::ObserverPtr<ObservedType>::ObserverPtr(ObservedType* const observe) :
	observed_{observe}
{}

template<typename ObservedType>
drn::foundation::ObserverPtr<ObservedType>::ObserverPtr(
	const std::unique_ptr<ObservedType>& observe
) :
	ObserverPtr{observe.get()}
{}

template<typename ObservedType>
drn::foundation::ObserverPtr<ObservedType>::ObserverPtr(const UniqueQtPtr<ObservedType>& observe) :
	ObserverPtr{observe.get()}
{}

template<typename ObservedType>
drn::foundation::ObserverPtr<ObservedType>& drn::foundation::ObserverPtr<ObservedType>::operator=(
	ObservedType* const observe
)
{
	this->observed_ = observe;
	return *this;
}

template<typename ObservedType>
drn::foundation::ObserverPtr<ObservedType>& drn::foundation::ObserverPtr<ObservedType>::operator=(
	const std::nullptr_t
)
{
	this->observed_ = nullptr;
	return *this;
}

template<typename ObservedType>
bool drn::foundation::ObserverPtr<ObservedType>::operator==(const std::nullptr_t) const
{
	return this->observed_ == nullptr;
}

template<typename ObservedType>
bool drn::foundation::ObserverPtr<ObservedType>::operator==(const ObserverPtr& other) const
{
	return this->observed_ == other.observed_;
}

template<typename ObservedType>
bool drn::foundation::ObserverPtr<ObservedType>::operator==(const ObservedType* observed) const
{
	return this->observed_ == observed;
}

template<typename ObservedType>
bool drn::foundation::ObserverPtr<ObservedType>::operator!=(const std::nullptr_t) const
{
	return ! (*this == nullptr);
}

template<typename ObservedType>
bool drn::foundation::ObserverPtr<ObservedType>::operator!=(const ObserverPtr& other) const
{
	return ! (*this == other);
}

template<typename ObservedType>
bool drn::foundation::ObserverPtr<ObservedType>::operator!=(const ObservedType* observed) const
{
	return ! (*this == observed);
}

template<typename ObservedType>
bool drn::foundation::ObserverPtr<ObservedType>::operator<(const ObservedType* observed) const
{
	return std::less<const ObservedType*>{}(this->observed_, observed);
}

template<typename ObservedType>
bool drn::foundation::ObserverPtr<ObservedType>::operator<(const ObserverPtr& other) const
{
	return *this < other.observed_;
}

template<typename ObservedType>
ObservedType* drn::foundation::ObserverPtr<ObservedType>::get() const noexcept
{
	return this->observed_;
}

template<typename ObservedType>
ObservedType& drn::foundation::ObserverPtr<ObservedType>::operator*()
{
	assert(this->observed_ != nullptr && "Should not access the pointed object when it's null.");
	return *this->observed_;
}

template<typename ObservedType>
ObservedType& drn::foundation::ObserverPtr<ObservedType>::operator*() const
{
	assert(this->observed_ != nullptr && "Should not access the pointed object when it's null.");
	return *this->observed_;
}

template<typename ObservedType>
ObservedType* drn::foundation::ObserverPtr<ObservedType>::operator->()
{
	assert(this->observed_ != nullptr && "Should not access the pointed object when it's null.");
	return this->observed_;
}

template<typename ObservedType>
ObservedType* drn::foundation::ObserverPtr<ObservedType>::operator->() const
{
	assert(this->observed_ != nullptr && "Should not access the pointed object when it's null.");
	return this->observed_;
}

template<typename ObservedType>
drn::foundation::ObserverPtr<ObservedType>::operator bool() const noexcept
{
	return this->observed_ != nullptr;
}

template<typename ObservedType>
drn::foundation::ObserverPtr<ObservedType>::operator ObservedType*() const noexcept
{
	return this->observed_;
}

template<typename ObservedType>
bool drn::foundation::operator==(std::nullptr_t, const ObserverPtr<ObservedType>& rhs)
{
	return rhs == nullptr;
}

template<typename ObservedType, typename ObservedPtrType>
bool drn::foundation::operator==(const ObservedType* lhs, const ObserverPtr<ObservedPtrType>& rhs)
{
	return lhs == rhs.get();
}

template<typename ObservedType>
bool drn::foundation::operator!=(std::nullptr_t, const ObserverPtr<ObservedType>& rhs)
{
	return ! (nullptr == rhs);
}

template<typename ObservedType, typename ObservedPtrType>
bool drn::foundation::operator!=(const ObservedType* lhs, const ObserverPtr<ObservedPtrType>& rhs)
{
	return ! (lhs == rhs);
}

template<typename ObservedType>
std::ostream& drn::foundation::operator<<(
	std::ostream& o,
	const ObserverPtr<ObservedType>& observer
)
{
	if (observer == nullptr)
		return o << "<nullptr>";
	return o << *observer;
}

#endif
