/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_FOUNDATION_PERCENTAGE_H_
#define DRN_FOUNDATION_PERCENTAGE_H_

#include <iosfwd>

#include <QtCore>

#include <foundation_export.h>


class QDebug;
class QString;

namespace drn::foundation
{

/**
 * @brief A representation of a percentage rate.
 */
class DRN_FOUNDATION_EXPORT Percentage
{
	friend DRN_FOUNDATION_EXPORT std::ostream& operator<<(std::ostream&, const Percentage&);
	friend DRN_FOUNDATION_EXPORT ::QString presentationText(const Percentage&);

	quint8 precision_;
	double rate_;

public:
	// N.B. The number of decimal places is chosen because it seems to be good enough for the number
	// of digits most percentage rates are advertised in.
	static constexpr const quint8 defaultPrecision{6u};

	/**
	 * @brief Full initialisation constructor.
	 *
	 * @exception Error When the rate is too large to truncate.
	 *
	 * @param rate The percentage rate value. Any digits after the precision is truncated.
	 * @param precision The number of digits after the decimal point.
	 */
	explicit Percentage(
		const double rate = 0.0,
		const quint8 precision = Percentage::defaultPrecision
	);
	/**
	 * @brief Conversion operator to the decimal representation of the percentage.
	 *
	 * @return The current percentage rate.
	 */
	explicit operator double() const noexcept;
	/// @brief An accessor to the number of digits to use in any operation involving the rate.
	[[nodiscard]]
	quint8 precision() const noexcept;
	[[nodiscard]]
	double rate() const noexcept;
	/**
	 * @brief Determines if one rate is less than the another using the largest precision.
	 *
	 * @param other The rate to determine if the current object's value is less-than.
	 *
	 * @return True when the two rates are less-than the largest epsilon digit places of each other,
	 * false else wise.
	 */
	[[nodiscard]]
	bool operator<(const Percentage& other) const;
	/**
	 * @brief Determines if one rate is less than the another.
	 *
	 * @param rate The rate to determine if the current object's value is less-than up to the
	 * default precision.
	 *
	 * @return True when the two rates are less-than the largest epsilon digit places of each other,
	 * false else wise.
	 */
	[[nodiscard]]
	bool operator<(const double& rate) const;
	/**
	 * @brief Determines if one rate is greater than the another using the largest precision.
	 *
	 * @param other The rate to determine if the current object's value is greater-than.
	 *
	 * @return True when the two rates are greater-than the largest epsilon digit places of each
	 * other, false else wise.
	 */
	[[nodiscard]]
	bool operator>(const Percentage& other) const;
	/**
	 * @brief Determines if one rate is greater than the another.
	 *
	 * @param rate The rate to determine if the current object's value is greater-than up to the
	 * default precision.
	 *
	 * @return True when the two rates are greater-than the largest epsilon digit places of each
	 * other, false else wise.
	 */
	[[nodiscard]]
	bool operator>(const double& rate) const;
	/**
	 * @brief Determines if two rates are the same.
	 *
	 * @param other The other rate to compare.
	 *
	 * @return True when the two rates are within the epsilon digit places of each other, false
	 * else wise.
	 */
	[[nodiscard]]
	bool operator==(const Percentage& other) const;
	/**
	 * @brief Determines if two rates are not the same.
	 *
	 * @param other The other rate to compare.
	 *
	 * @return True when the two rates are not within the epsilon digit places of each other, false
	 * else wise.
	 */
	[[nodiscard]]
	bool operator!=(const Percentage& other) const;
};

DRN_FOUNDATION_EXPORT std::ostream& operator<<(std::ostream& out, const Percentage& p);
DRN_FOUNDATION_EXPORT ::QDebug& operator<<(::QDebug& out, const Percentage& p);
DRN_FOUNDATION_EXPORT ::QString presentationText(const Percentage& p);

}

#endif
