/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_NAVIGATION_NAVIGATOR_H_
#define DRN_NAVIGATION_NAVIGATOR_H_

#include <pecunia/Money.h>

#include <QObject>
#include <QString>

#include <chrono>
#include <map>
#include <memory>
#include <optional>
#include <set>
#include <vector>

#include <accounting/Account.h>
#include <accounting/AccountCode.h>
#include <accounting/LedgerBooks.h>
#include <accounting/TransactionNumber.h>
#include <adviser/BudgetHealth.h>
#include <banking/BankAccount.h>
#include <banking/BankAccountType.h>
#include <banking/BankMap.h>
#include <banking/BankName.h>
#include <budgeting/BillMap.hpp>
#include <budgeting/Budget.h>
#include <budgeting/BudgetItemDue.h>
#include <budgeting/BudgetItemIdentifier.h>
#include <budgeting/BudgetItemTypes.h>
#include <budgeting/BudgetSource.h>
#include <budgeting/BudgetedMoney.h>
#include <budgeting/DebtMap.h>
#include <budgeting/GoalMap.h>
#include <budgeting/NontrackMap.hpp>
#include <budgeting/WageMap.hpp>
#include <conversion/ConversionMap.h>
#include <conversion/CurrencySource.hpp>
#include <foundation/Percentage.h>
#include <foundation/QtMemory.hpp>
#include <foundation/Switch.h>
#include <navigation_export.h>
#include <networking/RequestManager.hpp>
#include <surveying/IndexedIncreaseDecreases.hpp>
#include <surveying/MappingSurvey.h>
#include <updating/ReleaseInformation.hpp>
#include <updating/Version.hpp>

#include "BudgetBankAccount.h"
#include "UserPreferences.h"


namespace drn::storage
{

struct BudgetStorage;
struct ExchangeRatesStorage;

}
namespace drn::navigation::internal
{

class BudgetBankLedgers;

}
namespace drn::navigation
{

struct AllowSubAccounts;
enum class CurrencyDisplayOptions;

class DRN_NAVIGATION_EXPORT Navigator final : public ::QObject
{
	Q_OBJECT

	void markCleared(
		const accounting::TransactionNumber& txnNum,
		const std::optional<drn::accounting::AccountNumber>& actNum
	);
	void markUnreconciled(const accounting::TransactionNumber& tn);

	std::optional<::QString> filePathName_;
	foundation::Switch isDirty_;
	std::unique_ptr<internal::BudgetBankLedgers> budgetBankLedgers_;
	UserPreferences preferences_;
	foundation::QtPtr<networking::RequestManager> requestManager_;
	foundation::QtPtr<conversion::CurrencySource> exchangeSource_;
	std::unique_ptr<storage::BudgetStorage> budgetStore_;
	std::unique_ptr<storage::ExchangeRatesStorage> exchangeRatesStore_;

public slots:
	void onPrepareShowBills();
	void onPrepareShowDebts();
	void onPrepareShowGoals();
	void onPrepareShowNontracks();
	void onPrepareShowWages();
	void onAdded(
		const drn::budgeting::BudgetItemTypes& type,
		const drn::budgeting::BudgetedMoney& item,
		const std::optional<drn::banking::BankAccount>& ba
	);
	void onRenamed(
		const drn::budgeting::BudgetItemTypes& type,
		const drn::budgeting::BudgetSource& original,
		const drn::budgeting::BudgetSource& changed
	);
	void onUpdated(
		const drn::budgeting::BudgetItemTypes& type,
		const drn::budgeting::BudgetedMoney& item,
		const std::optional<drn::banking::BankAccount>& ba
	);
	void onRemoved(
		const drn::budgeting::BudgetItemTypes& type,
		const drn::budgeting::BudgetSource& source
	);
	void onPrepareShowAccounts();
	void onAddedAct(const drn::accounting::Account& act);
	void onRecoded(
		const drn::accounting::AccountCode& current,
		const drn::accounting::AccountCode& changed
	);
	void onClosed(const drn::accounting::AccountCode& code);
	void onOpened(const drn::accounting::AccountCode& code);
	void onRemovedAct(const drn::accounting::AccountCode& code);
	void onPrepareShowLedgers();
	void onSelectedLedger(const drn::accounting::AccountCode& code);
	void onAddedTransaction(const drn::accounting::Transaction& txn);
	void onUpdatedTransaction(
		const drn::accounting::TransactionNumber& tn,
		const drn::accounting::Transaction& txn
	);
	void onMarkCleared(const std::set<drn::accounting::TransactionNumber>& tns);
	void onMarkUnreconciled(const std::set<drn::accounting::TransactionNumber>& tns);
	void onMarkReconciled(const drn::accounting::TransactionNumber& tn);
	void onRemovedTransaction(const drn::accounting::TransactionNumber& tn);
	void onPrepareShowBanks();
	void onBankAdded(
		const drn::banking::Bank& b,
		const std::map<
			drn::accounting::AccountNumber,
			drn::banking::SupportedAccountTypes
		>& accountNumbersBankTypes
	);
	void onBankUpdated(
		const drn::banking::Bank& b,
		const std::map<
			drn::accounting::AccountNumber,
			drn::banking::SupportedAccountTypes
		>& accountNumbersBankTypes
	);
	void onBankRenamed(
		const drn::banking::BankName& original,
		const drn::banking::BankName& change
	);
	void onBankClosed(const drn::banking::BankName& bn);
	void onBankOpened(const drn::banking::BankName& bn);
	void onBankRemoved(const drn::banking::BankName& bn);
	void onAddedBankAccount(
		const QString& name,
		const drn::banking::SupportedAccountTypes& type,
		const pecunia::currency::Money& openingBalance
	);
	void onRenamedBankAccount(
		const drn::accounting::AccountCode& original,
		const drn::accounting::AccountCode& changed
	);
	void onRemovedBankAccount(const drn::accounting::AccountCode& ac);
	void onPrepareShowPostBudgetItems();
	void onPrepareShowSurvey();
	void onLookUpBankAccounts(
		const drn::banking::BankName& bn,
		const drn::navigation::AllowSubAccounts& shouldAllow
	);
	void onLookUpBudgetItemSources(
		const drn::budgeting::BudgetItemDueIdentifier& requestedBy,
		const drn::budgeting::BudgetItemTypes& bit
	);
	void onPostBudgetItemPosted(
		const drn::budgeting::BudgetItemAmountDue& bid,
		const drn::accounting::AccountCode& ac,
		const QDate& postOn
	);
	void onPostBudgetItemPostedUnexpected(
		const drn::budgeting::BudgetItemDueIdentifier& bidi,
		const pecunia::currency::Money& amount,
		const drn::accounting::AccountCode& ac,
		const QDate& postOn
	);
	void onPostBudgetItemSkipped(const drn::budgeting::BudgetItemDueIdentifier& bidi);
	void onNewBudgetFile();
	void onConfirmedNewBudgetFile();
	void onLoadBudgetFile(const QString& filePathName);
	void onConfirmedLoadBudgetFile(const QString& filePathName, const bool isReloaded);
	void onSaveBudgetFile();
	void onSaveAsBudgetFile(const QString& filePathName);
	void onReloadBudgetFile();
	void onCheckIsBudgetFileDirty();
	void onPrepareShowDashboard();
	void onPrepareShowExchangeRates();
	void onFetchExchangeRates(const std::set<pecunia::currency::Currency>& currencies);
	void onStoreExchangeRates(const drn::conversion::ConversionMap& conversions);
	void onPrepareShowConfigure();
	void onConfigureUpdated(
		const std::chrono::hours& hrInWeek,
		const std::chrono::days& dayInWeek,
		const std::chrono::days& notifyDueWithin,
		const pecunia::currency::Currency& preferred,
		const drn::navigation::CurrencyDisplayOptions& displayAs,
		const std::set<pecunia::currency::Currency>& usable,
		const drn::foundation::Percentage& balanceWindow
	);
	void onLookUpAccountCurrency(const drn::accounting::AccountNumber& an);
	void onGenerateMappingSurvey(
		const drn::banking::BankName& bn,
		const drn::accounting::AccountNumber& an,
		const QDate& surveyingOn
	);
	void onCompletedSurvey(
		const QDate& completedOn,
		const drn::banking::BankAccount& ba,
		const pecunia::currency::Money& balance,
		const std::set<drn::accounting::TransactionNumber>& allReconciled,
		const drn::surveying::IndexedIncreaseDecreases<
			drn::budgeting::BudgetItemIdentifier
		>& allAdded,
		const drn::surveying::IndexedIncreaseDecreases<drn::accounting::AccountNumber>& allExtra,
		const std::map<drn::budgeting::BudgetItemIdentifier, pecunia::currency::Money>& distributed,
		const pecunia::currency::Money& nonDistributed
	);
	void onFetchReleaseNotes();
	void onCheckForUpdate();

signals:
	void showBills(
		const drn::budgeting::BillMap& bills,
		const std::chrono::hours& hrInWeek,
		const std::chrono::days& dayInWeek,
		const pecunia::currency::Currency& preferred,
		const std::set<pecunia::currency::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs
	);
	void showDebts(
		const drn::budgeting::DebtMap& debts,
		const std::set<drn::navigation::BudgetBankAccount>& accounts,
		const std::set<drn::accounting::AccountCode>& codes,
		const std::set<drn::banking::BankName>& banks,
		const std::chrono::hours& hrInWeek,
		const std::chrono::days& dayInWeek,
		const pecunia::currency::Currency& preferred,
		const std::set<pecunia::currency::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs
	);
	void showGoals(
		const drn::budgeting::GoalMap& goals,
		const std::set<drn::navigation::BudgetBankAccount>& accounts,
		const std::set<drn::accounting::AccountCode>& codes,
		const std::set<drn::banking::BankName>& banks,
		const std::chrono::hours& hrInWeek,
		const std::chrono::days& dayInWeek,
		const pecunia::currency::Currency& preferred,
		const std::set<pecunia::currency::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs,
		const pecunia::currency::Money& wagesTotal
	);
	void showNontracks(
		const drn::budgeting::NontrackMap& nontracks,
		const std::chrono::hours& hrInWeek,
		const std::chrono::days& dayInWeek,
		const pecunia::currency::Currency& preferred,
		const std::set<pecunia::currency::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs
	);
	void showWages(
		const drn::budgeting::WageMap& wages,
		const std::chrono::hours& hrInWeek,
		const std::chrono::days& dayInWeek,
		const pecunia::currency::Currency& preferred,
		const std::set<pecunia::currency::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs
	);
	void saveSucceeded(const drn::budgeting::BudgetSource& source);
	void saveFailed(
		const std::optional<drn::budgeting::BudgetSource>& original,
		const QString& errorMessage,
		const std::optional<drn::budgeting::BudgetSource>& displayed
	);
	void showAccounts(
		const std::set<drn::accounting::Account>& accounts,
		const std::set<pecunia::currency::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs
	);
	void accountSaveSucceeded(const drn::accounting::AccountCode& code);
	void accountSaveFailed(
		const std::optional<drn::accounting::AccountCode>& original,
		const QString& errorMessage,
		const std::optional<drn::accounting::AccountCode>& displayed
	);
	void showLedgers(
		const std::set<drn::accounting::Account>& accounts,
		const std::set<pecunia::currency::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs
	);
	void showLedger(const drn::accounting::LedgerAccount& ledger);
	void transactionSaveSucceeded(
		const std::optional<drn::accounting::TransactionNumber>& displayed,
		const std::optional<drn::accounting::TransactionNumber>& posted
	);
	void transactionSaveFailed(
		const std::optional<drn::accounting::TransactionNumber>& tn,
		const QString& errorMessage
	);
	void showBanks(
		const drn::banking::BankMap& banks,
		const std::set<drn::accounting::Account>& allAccounts,
		const std::set<pecunia::currency::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs,
		const std::set<drn::banking::BankAccountType>& accountTypes
	);
	void bankSaveSucceeded(const drn::banking::BankName& bn);
	void bankSaveFailed(
		const std::optional<drn::banking::BankName>& original,
		const QString& errorMessage,
		const std::optional<drn::banking::BankName>& displayed
	);
	void bankAccountSaveSucceeded(const drn::accounting::AccountCode& ac);
	void bankAccountSaveFailed(
		const std::optional<drn::accounting::AccountCode>& original,
		const QString& errorMessage,
		const std::optional<drn::accounting::AccountCode>& displayed
	);
	void showPostBudgetItems(
		const std::chrono::days& dueWithin,
		const std::vector<drn::budgeting::BudgetItemAmountDue>& items,
		const std::set<drn::banking::BankName>& banks,
		const pecunia::currency::Currency& preferredCurrency,
		const std::set<pecunia::currency::Currency>& usableCurrencies,
		const drn::navigation::CurrencyDisplayOptions& displayAs,
		const std::set<drn::accounting::AccountCode>& codes
	);
	void showSurvey(
		const std::set<drn::banking::BankName>& banks,
		const std::set<pecunia::currency::Currency>& usable,
		const drn::navigation::CurrencyDisplayOptions& displayAs,
		const drn::foundation::Percentage& balanceWindow,
		const std::set<drn::accounting::AccountCode>& codes
	);
	void lookedUpBankAccounts(
		const drn::banking::BankName& bn,
		const std::set<drn::banking::BankAccountType>& accountTypes
	);
	void lookedUpBudgetItemSources(
		const drn::budgeting::BudgetItemDueIdentifier& requestedBy,
		const drn::budgeting::BudgetItemTypes& bit,
		const std::set<drn::budgeting::BudgetSource>& sources
	);
	void postBudgetItemSaveSucceeded(
		const drn::budgeting::BudgetItemDueIdentifier& bidi,
		const std::optional<drn::accounting::TransactionNumber>& txnNum
	);
	void postBudgetItemSaveFailed(
		const drn::budgeting::BudgetItemDueIdentifier& bidi,
		const QString& errorMessage
	);
	void showConfirmNewBudgetFile();
	void clearedBudgetFile();
	void showConfirmLoadBudgetFile(const QString& filePathName);
	void loadedBudgetFile(const QString& filePathName, const bool isReloaded);
	void showSaveAsBudgetFile();
	void savedBudgetFile(const QString& filePathName);
	void showNoBudgetFileOpened();
	void checkedIsBudgetFileDirty(const drn::foundation::Switch& isDirty);
	void showDashboard(
		const drn::budgeting::BudgetTotals& bt,
		const std::vector<drn::budgeting::BudgetItemAmountDue>& due,
		const drn::adviser::BudgetHealth& bh,
		const std::chrono::days& dueWithin,
		const pecunia::currency::Currency& preferred,
		const drn::navigation::CurrencyDisplayOptions& displayAs
	);
	void showExchangeRates(
		const drn::conversion::ConversionMap& currentConversions,
		const std::set<pecunia::currency::Currency>& currencies
	);
	void fetchedExchangeRates(
		const std::set<pecunia::currency::Currency>& currencies,
		const drn::conversion::ConversionMap& conversions
	);
	void showConfigure(
		const std::chrono::hours& hrInWeek,
		const std::chrono::days& dayInWeek,
		const std::chrono::days& notifyDueWithin,
		const pecunia::currency::Currency& preferred,
		const drn::navigation::CurrencyDisplayOptions& displayAs,
		const std::set<pecunia::currency::Currency>& usable,
		const std::set<pecunia::currency::Currency>& used,
		const drn::foundation::Percentage& balanceWindow
	);
	void configureSaveSucceeded();
	void configureSaveFailed(const QString& errorMessage);
	void lookedUpAccountCurrency(const pecunia::currency::Currency& code);
	void generatedMappingSurvey(const drn::surveying::MappingSurvey& survey);
	void completedSurveySucceeded();
	void completedSurveyFailed(const QString& errorMessage);
	void fetchedReleaseNotes(
		const std::set<drn::updating::ReleaseInformation>& versions
	);
	void updateReleased(const drn::updating::SemanticVersion& latest);

public:
	Navigator(
		foundation::UniqueQtPtr<networking::RequestSender> sender,
		std::unique_ptr<storage::BudgetStorage> budgetStore,
		std::unique_ptr<storage::ExchangeRatesStorage> exchangeRatesStore
	);
	~Navigator() override;
};

}

#endif
