/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_NAVIGATION_USERPREFERENCES_H_
#define DRN_NAVIGATION_USERPREFERENCES_H_

#include <QSettings>

#include <chrono>
#include <set>

#include <navigation_export.h>


namespace pecunia::currency
{

enum class Currency : quint16;

}

namespace drn::foundation
{

class Percentage;

}
namespace drn::navigation
{

enum class CurrencyDisplayOptions;

/**
 * @brief The representation for the setting values the users prefers to use in the application.
 */
class DRN_NAVIGATION_EXPORT UserPreferences
{
	QSettings settings_;

public:
	UserPreferences();

	//{ Accounting Preferences

	/**
	 * @brief Provides the preference of the user for the survey balance window.
	 *
	 * @return The +/- percentage value a survey balance must be within to be considered balanced.
	 * When there is not a user preference value, a default is used instead.
	 */
	[[nodiscard]]
	foundation::Percentage surveyWindow() const noexcept;
	/**
	 * @brief Stores the user's preference for the survey balance window.
	 *
	 * @param window The value to store.
	 */
	void setSurveyWindow(const foundation::Percentage& window);

	//}

	//{ budgeting Preferences

	/**
	 * @brief Provides the preference of the user for the number of work hours in a week.
	 *
	 * @return The normal/average number of work hours in a week. When there is not a user
	 * preference value, a default is used instead.
	 */
	[[nodiscard]]
	std::chrono::hours hoursInWeek() const noexcept;
	/**
	 * @brief Stores the user's preference for the number of work hours in a week.
	 *
	 * @param value The value to store.
	 */
	void setHoursInWeek(const std::chrono::hours& value);
	/**
	 * @brief Provides the preference of the user for the number of work days in a week.
	 *
	 * @return The normal/average number of work days in a week. When there is not a user
	 * preference value, a default is used instead.
	 */
	[[nodiscard]]
	std::chrono::days daysInWeek() const noexcept;
	/**
	 * @brief Stores the user's preference for the number of work days in a week.
	 *
	 * @param value The value to store.
	 */
	void setDaysInWeek(const std::chrono::days& value);
	/**
	 * @brief Provides the preference of the user for the number of days to notify that a budget
	 * item is due.
	 *
	 * @return The number of days to notify. When there is not a user preference value, a default is
	 * used instead.
	 */
	[[nodiscard]]
	std::chrono::days notifyDueWithin() const noexcept;
	/**
	 * @brief Stores the user's preference for the number of days to notify that a budget
	 * item is due.
	 *
	 * @param value The value to store.
	 */
	void setNotifyDueWithin(const std::chrono::days& value);

	//}

	//{ currency Preferences

	/**
	 * @brief Provides the preference of the user for the preferred currency to display monthly sums
	 * in, etc.
	 *
	 * @return The currency to use for sums, etc. When there is not a user preference value, a
	 * default is used instead.
	 */
	[[nodiscard]]
	pecunia::currency::Currency preferredCurrency() const noexcept;
	/**
	 * @brief Stores the user's preference for the preferred currency to display monthly sums
	 * in, etc.
	 *
	 * @param value The value to store.
	 */
	void setPreferredCurrency(const pecunia::currency::Currency& value);
	/**
	 * @brief Provides the preference of the user for all the currencies he wants to work with.
	 *
	 * @return Every currency to work with in the application. When there is not a user preference
	 * value, a default is used instead.
	 */
	[[nodiscard]]
	std::set<pecunia::currency::Currency> usableCurrencies() const noexcept;
	/**
	 * @brief Stores the user's preference for all the currencies he wants to work with.
	 *
	 * @param value The values to store.
	 */
	void setUsableCurrencies(const std::set<pecunia::currency::Currency>& values);
	[[nodiscard]]
	/**
	 * @brief Provides the preference of the user for how currencies should be displayed.
	 *
	 * @return The preferred display method when set. When there is not a user preference the
	 * default of displaying the ISO-4217 code is used.
	 */
	navigation::CurrencyDisplayOptions currenciesDisplayAs() const noexcept;
	/**
	 * @brief Stores the user's preference for how currencies should be displayed.
	 *
	 * @param displayAs The value to store.
	 */
	void setCurrenciesDisplayAs(const navigation::CurrencyDisplayOptions& displayAs);

	//}
};

}

#endif
