/*******************************************************************************
***  This file is part of Dux Rei Nummariae.                                 ***
***                                                                          ***
***  Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024            ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Affero General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Affero General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Affero General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef DRN_SURVEYING_MAPPING_SURVEY_H_
#define DRN_SURVEYING_MAPPING_SURVEY_H_

#include <pecunia/Money.h>

#include <QDate>

#include <compare>
#include <map>
#include <set>

#include <accounting/AccountCode.h>
#include <accounting/Transaction.h>
#include <banking/Reconciliation.h>
#include <budgeting/BudgetItemIdentifier.h>
#include <surveying_export.h>


namespace drn::budgeting
{

enum class BudgetItemTypes : quint8;
enum class EventFrequency : quint8;

}
namespace drn::surveying
{

/**
 * @brief A representation of a balance change for an account.
 */
struct DRN_SURVEYING_EXPORT BalanceChange
{
	/**
	 * @brief Performs a comparison based on the transaction and isIncrease value.
	 *
	 * @param other The other object to compare with.
	 *
	 * @return Returns strong_ordering::less if this object's transaction is less than the other
	 * object, strong_ordering::equal if the transaction and increase flag are equal, and
	 * strong_ordering::greater if this object is greater than the other object.
	 */
	std::strong_ordering operator<=>(const BalanceChange& other) const;
	bool operator==(const BalanceChange&) const = default;

	/// A flag, when true, indicates that the transaction is a balance increase for the account.
	bool isIncrease_{};
	/// The transaction responsible for the account balance change.
	accounting::Transaction transaction_{};
};
/**
 * @brief A representation of a survey mapping before being completed.
 */
class DRN_SURVEYING_EXPORT MappingSurvey
{
	banking::ReconciledBankAccount reconciled_;
	pecunia::currency::Money expectedBalance_;
	::QDate calculatedFrom_;
	std::map<budgeting::BudgetItemIdentifier, std::set<BalanceChange>> used_;
	std::map<budgeting::BudgetItemIdentifier, pecunia::currency::Money> distributable_;
	std::map<accounting::AccountCode, std::set<BalanceChange>> additional_;

public:
	MappingSurvey() = default;
	/**
	 * @brief Full initialisation constructor.
	 *
	 * @exception BankError When the bank account and the expected balances have different
	 * currencies.
	 * @exception BankError When the expected balance calculation date is invalid.
	 *
	 * @param reconciled The bank account that the survey mapping is being performed upon
	 * @param expectedBalance The calculated expected balance of the bank account.
	 * @param calculatedFrom The date from which the expected balance is calculated from.
	 * @param used The budget items used since the last mapping.
	 * @param distributable The way the expected balance is expected to have been distributed.
	 * @param additional The extra transactions involving the bank account, but not a budget item.
	 */
	MappingSurvey(
		banking::ReconciledBankAccount reconciled,
		pecunia::currency::Money expectedBalance,
		::QDate calculatedFrom,
		std::map<budgeting::BudgetItemIdentifier, std::set<BalanceChange>> used,
		std::map<budgeting::BudgetItemIdentifier, pecunia::currency::Money> distributable,
		std::map<accounting::AccountCode, std::set<BalanceChange>> additional
	);
	/**
	 * @brief An accessor to the bank account that is being mapped for reconciliation.
	 */
	[[nodiscard]]
	const banking::ReconciledBankAccount& reconciled() const noexcept;
	/**
	 * @brief An accessor to the balance of the bank account calculated from the budget information.
	 * N.B. this may, and often is, different that the reconciled balance from a bank statement.
	 */
	[[nodiscard]]
	const pecunia::currency::Money& expectedBalance() const noexcept;
	/**
	 * @brief An accessor to the date the expected balance of the bank account is calculated from.
	 */
	[[nodiscard]]
	const ::QDate& calculatedFrom() const noexcept;
	/**
	 * @brief An accessor to all the budget item transactions that are not reconciled for the bank
	 * account.
	 */
	[[nodiscard]]
	const std::map<budgeting::BudgetItemIdentifier, std::set<BalanceChange>>& used() const noexcept;
	/**
	 * @brief An accessor to the predicted way the expected balance could be distributable
	 * throughout the budget.
	 */
	[[nodiscard]]
	const std::map<budgeting::BudgetItemIdentifier, pecunia::currency::Money>& distributable(
	) const noexcept;
	/**
	 * @brief An accessor to any additional transactions involving the bank account but not a budget
	 * item, e.g. an opening balance, paying budget items using child accounts, etc.
	 */
	[[nodiscard]]
	const std::map<accounting::AccountCode, std::set<BalanceChange>>& additional() const noexcept;
};

/**
 * @brief Determines if a budget item is distributable in a survey.
 *
 * @param type The type of the budget item.
 * @param period The frequency at which the budget item occurs.
 *
 * @return Gives true when the budget item is distributable, otherwise false.
 */
DRN_SURVEYING_EXPORT bool isDistributable(
	const budgeting::BudgetItemTypes& type,
	const budgeting::EventFrequency& period
);

}

#endif
